/*
COPYRIGHT, LICENSE AND WARRANTY INFORMATION

This software module has been originally developed by Nokia Corporation. 

The software module is provided free of charge and made available for use 
by the Licensee, who shall be a member of the MPEG standardization 
committee, under the following conditions:

The Supplied Data may only be used for the purpose of developing, testing 
and promulgating the MPEG technology standard(s) for 3D video coding.

Nokia retains the ownership of copyrights to the Software. There is no 
patent nor other intellectual property right of Nokia licensed under this 
Statement (except the copyright license above). Licensee hereby assumes 
sole responsibility to secure any other intellectual property rights 
needed, if any. For example, if patent licenses  are required, it is 
their responsibility to acquire the license before utilizing the Software.

The license by Nokia is subject to that the Licensee grants to Nokia the 
non-exclusive, worldwide, royalty-free, perpetual and irrevocable 
covenant that the Licensee(s) shall not bring a suit before any court or 
administrative agency or otherwise assert a claim for infringement under 
the Licensee intellectual property rights that, but for a license, would 
be infringed by the Software against  
   (a)  Nokia or Nokia's Affiliate; or  
   (b)  other recipient of a license and covenant not to sue with respect 
        to the Software from Nokia; or 
   (c)  contractor, customer or distributor of a party listed above in a 
        or b,  
which suit or claim is related to the Software or use thereof.

The Licensee(s) further agrees to grant a reciprocal license to Nokia 
(as granted by Nokia to the Licensee(s) on the modifications made by 
Licensee(s) to the Software. 

THE SOFTWARE IS PROVIDED "AS IS" AND THE ORIGINAL DEVELOPER DISCLAIMS 
ANY AND ALL WARRANTIES WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING 
BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS 
FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. THOSE INTENDING TO USE 
THE SOFTWARE ARE EXPRESSLY ADVISED THAT ITS USE MAY INFRINGE EXISTING 
PATENTS AND BE SUBJECT TO ROYALTY PAYMENTS TO PATENT OWNERS. ANYONE 
USING THE SOFTWARE ON THE BASIS OF THIS LICENSE AGREES TO OBTAIN THE 
NECESSARY PERMISSIONS FROM ANY AND ALL APPLICABLE PATENT OWNERS FOR SUCH 
USE.

IN NO EVENT SHALL THE ORIGINAL DEVELOPER BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

This copyright, license and warranty information notice must be retained 
in all copies and derivative works of the Software or substantial 
portions thereof.
*/

/*!
 *************************************************************************************
 * \file mc_depth_prediction.c
 *
 * \brief
 *    interpolate the 1/4-pixel accurate depth map with bi-linear filter 
 *
 * \author
 *      - Wenyi Su         <suwy@mail.ustc.edu.cn>
 *      - Dmytro           <Dmytro.Rusanovskyy@nokia.com>
 *
 *************************************************************************************
*/

#include "global.h"
#include "mc_depth_prediction.h"

/*!
************************************************************************
* \brief
*    Integer positions
************************************************************************
*/ 
static void get_depth_block_00(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos )
{
	// fastest to just move an entire block, since block is a temp block is a 256 byte block (16x16)
	// writes 2 lines of 16 imgpel 1 to 8 times depending in ver_block_size
	int j;
	imgpel *orig_line,*cur_line;

	for (j = 0; j < ver_block_size; ++j)
	{ 
		cur_line=&(cur_imgY[j][x_pos]);
		orig_line=block[j];
		memcpy(orig_line,cur_line,hor_block_size*sizeof(imgpel));
	}
}


/*!
************************************************************************
* \brief
*    Qpel (1,0) horizontal
************************************************************************
*/ 
static void get_depth_luma_10(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos , int max_imgpel_value)
{
	imgpel *p0, *p1;
	imgpel *orig_line;
	int i, j;
	int result;

	for (j = 0; j < ver_block_size; ++j)
	{
		p0 = &cur_imgY[j][x_pos];
		p1 = p0 + 1;

		orig_line = block[j];            

		for (i = 0; i < hor_block_size; ++i)
		{  

			result  = 3*(*p0++) +(*p1++);
			*orig_line = (imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Half horizontal
************************************************************************
*/ 
static void get_depth_luma_20(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos , int max_imgpel_value)
{
	imgpel *p0, *p1;
	imgpel *orig_line;
	int i, j;
	int result;
	for (j = 0; j < ver_block_size; j++)
	{
		p0 = &cur_imgY[j][x_pos];
		p1 = p0 + 1;
		orig_line = block[j];

		for (i = 0; i < hor_block_size; i++)
		{        
			result  = *(p0++) + *(p1++);

			*orig_line++ = (imgpel) iClip1(max_imgpel_value, ((result + 1)>>1));
		}
	}
}

/*!
************************************************************************
* \brief
*    Qpel (3,0) horizontal
************************************************************************
*/ 
static void get_depth_luma_30(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos , int max_imgpel_value)
{
	imgpel *p0, *p1;
	imgpel *orig_line;
	int i, j;
	int result;

	for (j = 0; j < ver_block_size; j++)
	{
		p0 = &cur_imgY[j][x_pos];
		p1 = p0 + 1;

		orig_line = block[j];            

		for (i = 0; i < hor_block_size; i++)
		{        
			result  = *(p0++) + 3*(*(p1++)) ;

			*orig_line = (imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Qpel vertical (0, 1)
************************************************************************
*/ 
static void get_depth_luma_01(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	imgpel *p0, *p1;
	imgpel *orig_line, *cur_line;
	int i, j;
	int result;
	int jj = 0;
	p0 = &(cur_imgY[0][x_pos]);
	for (j = 0; j < ver_block_size; j++)
	{                  
		p1 = p0 + shift_x;          

		orig_line = block[j];
		cur_line = &(cur_imgY[jj++][x_pos]);

		for (i = 0; i < hor_block_size; i++)
		{
			result  = 3*(*(p0++)) + *(p1++);

			*orig_line = (imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
		p0 = p1 - hor_block_size;
	}
}


/*!
************************************************************************
* \brief
*    Half vertical
************************************************************************
*/ 
static void get_depth_luma_02(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	imgpel *p0, *p1;
	imgpel *orig_line;
	int i, j;
	int result;
	p0 = &(cur_imgY[ 0][x_pos]);
	for (j = 0; j < ver_block_size; j++)
	{                  
		p1 = p0 + shift_x;          
		orig_line = block[j];

		for (i = 0; i < hor_block_size; i++)
		{
			result  = *(p0++) + *(p1++);

			*orig_line++ = (imgpel) iClip1(max_imgpel_value, ((result + 1)>>1));
		}
		p0 = p1 - hor_block_size;
	}
}


/*!
************************************************************************
* \brief
*    Qpel vertical (0, 3)
************************************************************************
*/ 
static void get_depth_luma_03(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	imgpel *p0, *p1;
	imgpel *orig_line;
	int i, j;
	int result;

	p0 = &(cur_imgY[ 0][x_pos]);
	for (j = 0; j < ver_block_size; j++)
	{                  
		p1 = p0 + shift_x;          

		orig_line = block[j];

		for (i = 0; i < hor_block_size; i++)
		{
			result  = *(p0++) + 3*(*(p1++)) ;

			*orig_line = (imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
		p0 = p1 - hor_block_size;
	}
}

/*!
************************************************************************
* \brief
*    Hpel horizontal, Qpel vertical (2, 1)
************************************************************************
*/ 
static void get_depth_luma_21(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=3*(*(p0++)+*(p1++))+*(p2++)+*(p3++);
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 4)>>3));
			orig_line++;
		}
	}

}

/*!
************************************************************************
* \brief
*    Hpel horizontal, Hpel vertical (2, 2)
************************************************************************
*/ 
static void get_depth_luma_22(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=*(p0++)+*(p1++)+*(p2++)+*(p3++);
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Hpel horizontal, Qpel vertical (2, 3)
************************************************************************
*/ 
static void get_depth_luma_23(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=*(p0++)+*(p1++)+3*(*(p2++)+*(p3++));
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 4)>>3));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Qpel horizontal, Hpel vertical (1, 2)
************************************************************************
*/ 
static void get_depth_luma_12(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=3*(*(p0++)+*(p2++))+*(p1++)+*(p3++);
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 4)>>3));
			orig_line++;
		}
	}
}


/*!
************************************************************************
* \brief
*    Qpel horizontal, Hpel vertical (3, 2)
************************************************************************
*/ 
static void get_depth_luma_32(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=3*(*(p1++)+*(p3++))+*(p0++)+*(p2++);
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 4)>>3));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Qpel horizontal, Qpel vertical (3, 3)
************************************************************************
*/ 
static void get_depth_luma_33(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=*(p1++)+*(p2++)+2*(*(p3++));
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}



/*!
************************************************************************
* \brief
*    Qpel horizontal, Qpel vertical (1, 1)
************************************************************************
*/ 
static void get_depth_luma_11(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=2*(*(p0++))+*(p1++)+*(p2++);
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Qpel horizontal, Qpel vertical (1, 3)
************************************************************************
*/ 
static void get_depth_luma_13(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=*(p0++)+*(p3++)+2*(*(p2++));
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}

/*!
************************************************************************
* \brief
*    Qpel horizontal, Qpel vertical (3, 1)
************************************************************************
*/ 
static void get_depth_luma_31(imgpel **block, imgpel **cur_imgY, int ver_block_size, int hor_block_size, int x_pos, int shift_x, int max_imgpel_value)
{
	int i, j;
	/* Vertical & horizontal interpolation */
	imgpel *p0, *p1, *p2, *p3;
	imgpel *orig_line;  
	int result;      


	for(j=0;j<ver_block_size;++j)
	{
		p0=&cur_imgY[j][x_pos];
		p1=&cur_imgY[j+1][x_pos];
		p2=p0+1;
		p3=p1+1;

		orig_line = block[j];
		for(i=0;i<hor_block_size;++i)
		{
			result=*(p0++)+*(p2++)+2*(*(p1++));
			*orig_line=(imgpel) iClip1(max_imgpel_value, ((result + 2)>>2));
			orig_line++;
		}
	}
}



/*!
 ************************************************************************
 * \brief
 *    Interpolation of 1/4 subpixel with bi-linear interpolation
 ************************************************************************
 */ 
void get_depth_mc_luma(imgpel** img_ref_Y, int x_pos, int y_pos, int hor_block_size, int ver_block_size, imgpel **block,
					int shift_x, int max_x, int max_y, int max_imgpel_value)
{
#if NOKIA_USTC_B_VSP_M26069
  const int int_pel_dis=1;
  int dx=0,dy=0;
#else
  int dx = (x_pos & 3);
  int dy = (y_pos & 3);
#if D_MVP_MODE
  const int int_pel_dis=1;
  x_pos=((x_pos+2) >>2);
  y_pos=((y_pos +2)>> 2);
#else
  x_pos >>= 2;
  y_pos >>= 2;
#endif //#if D_MVP_MODE
#endif//#if NOKIA_USTC_B_VSP_M26069

	x_pos = iClip3(-18, max_x+2, x_pos);
	y_pos = iClip3(-10, max_y+2, y_pos);

#if D_MVP_MODE
	if ((dx == 0 && dy == 0)||(int_pel_dis))
#else
	if (dx == 0 && dy == 0)
#endif
	{
		get_depth_block_00(block, &img_ref_Y[y_pos], ver_block_size, hor_block_size,x_pos);
	}
	else
	{ /* other positions */
		if (dy == 0) /* No vertical interpolation */
		{         
			if (dx == 1)
				get_depth_luma_10(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, max_imgpel_value);
			else if (dx == 2)
				get_depth_luma_20(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, max_imgpel_value);
			else
				get_depth_luma_30(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, max_imgpel_value);
		}
		else if (dx == 0) /* No horizontal interpolation */        
		{         
			if (dy == 1)
				get_depth_luma_01(block, &img_ref_Y[y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
			else if (dy == 2)
				get_depth_luma_02(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
			else
				get_depth_luma_03(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
		}
		else if (dx == 2)  /* Vertical & horizontal interpolation */
		{  
			if (dy == 1)
				get_depth_luma_21(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, max_imgpel_value);
			else if (dy == 2)
				get_depth_luma_22(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, max_imgpel_value);
			else
				get_depth_luma_23(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, max_imgpel_value);
		}
		else if (dy == 2)
		{
			if (dx == 1)
				get_depth_luma_12(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
			else
				get_depth_luma_32(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
		}
		else
		{
			if (dx == 1)
			{
				if (dy == 1)
					get_depth_luma_11(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
				else
					get_depth_luma_13(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
			}
			else
			{
				if (dy == 1)
					get_depth_luma_31(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
				else
					get_depth_luma_33(block, &img_ref_Y[ y_pos], ver_block_size, hor_block_size, x_pos, shift_x, max_imgpel_value);
			}
		}
	}
}