/*!
*************************************************************************************
* \file smrc_tmp_funcs.c
*
* \brief
*    Related functions for smrc tools
*
* \author
*    Main contributors
*     - Michael Mishourovsky   <mike.mishky@samsung.com>
*     - Alexis Fartukov        <a.fartukov@samsung.com>
*     - Igor Kovliga           <i.kovliga@samsung.com>
*     - Konstantin Dubkov      <k.dubkov@samsung.com>
*     
*************************************************************************************
*/
#include "contributors.h"

#include <math.h>

#ifdef _WIN32
#include <windows.h>
#include <crtdbg.h>
#endif // _WIN32

#include "global.h"
#include "input.h"
#include "img_io.h"
#include "memalloc.h"
#include "configfile.h"
#include "rdopt.h"

#include "image.h"
#include "mbuffer.h"
#include "transform.h"
#include "blk_prediction.h"
#include "md_common.h"

#if(DEBUG_PROCESSING)

#define BMP_LOG 1
#define THRESHOLD 2

#define NUM_SKIP_ZONE_MAP			0
#define NUM_VSPSKIP_ZONE_MAP	1
#define NUM_ALC_ZONE_MAP			2
#define NUM_ALC_PRED					3

#define QUANTITY_PARTITIONS		6

#define NUM_SKIP_ALC_PRED			(NUM_ALC_PRED+0)
#define NUM_16x16_ALC_PRED		(NUM_ALC_PRED+1)
#define NUM_16x8_ALC_PRED			(NUM_ALC_PRED+2)
#define NUM_8x16_ALC_PRED			(NUM_ALC_PRED+3)
#define NUM_8x8_ALC_PRED			(NUM_ALC_PRED+4)
#define NUM_MERL_ALC_PRED			(NUM_ALC_PRED+5)

#define NUM_SKIP_PURE_PRED		(QUANTITY_PARTITIONS+NUM_ALC_PRED+0)
#define NUM_16x16_PURE_PRED		(QUANTITY_PARTITIONS+NUM_ALC_PRED+1)
#define NUM_16x8_PURE_PRED		(QUANTITY_PARTITIONS+NUM_ALC_PRED+2)
#define NUM_8x16_PURE_PRED		(QUANTITY_PARTITIONS+NUM_ALC_PRED+3)
#define NUM_8x8_PURE_PRED			(QUANTITY_PARTITIONS+NUM_ALC_PRED+4)
#define NUM_MERL_PURE_PRED		(QUANTITY_PARTITIONS+NUM_ALC_PRED+5)


#define NUM_DEBUG_PICTURES (NUM_ALC_PRED+1 + 2*QUANTITY_PARTITIONS)

imgpel** Pictures[NUM_DEBUG_PICTURES]={0};
int frame_enable_flag;
imgpel** MskBlock255;
imgpel** MskBlock128;
imgpel** pR;
imgpel** pG;
imgpel** pB;

void FillRectangle( imgpel** pR, imgpel** pG, imgpel** pB, int height,int width,
                    int s_x, int s_y, int dx, int dy, imgpel R, imgpel G, imgpel B);
void GetPSNRwith2mask( imgpel** pPelOrig, imgpel** pPelRec, imgpel** pMaskMap1, imgpel** pMaskMap2, imgpel sig1, imgpel sig2, int h, int w, double* psnr, int* ProcessedNumber);
void GetMR4x4PSNRwith2mask( imgpel** pPelOrig, imgpel** pPelRec, imgpel** pMaskMap1, imgpel** pMaskMap2, imgpel sig1, imgpel sig2, int h, int w, double* psnr, int* ProcessedNumber);
int save_pict_imgpel(const char* filename, imgpel **pRImage, imgpel **pGImage, imgpel **pBImage, int height, int width);
void db_set_diff_frame( imgpel** pFrameSrc, imgpel** pLuma1, imgpel** pLuma2, int h, int w);

int debug_processing_init( VideoParameters* p_Vid)
{
	int i;
	frame_enable_flag = 0;
	get_mem2Dpel( &MskBlock128, 16, 16);
	get_mem2Dpel( &MskBlock255, 16, 16);
	memset( &(MskBlock128[0][0]), 128, 256);
	memset( &(MskBlock255[0][0]), 255, 256);
	for( i = 0; i < NUM_DEBUG_PICTURES;i++)
	{
		if( 0 == Pictures[i])
			get_mem2Dpel( Pictures+i, p_Vid->height, p_Vid->width);
	}

	get_mem2Dpel( &pR, p_Vid->height, p_Vid->width);
	get_mem2Dpel( &pG, p_Vid->height, p_Vid->width);
	get_mem2Dpel( &pB, p_Vid->height, p_Vid->width);

	return 0;
}

void debug_processing_deinit( VideoParameters* p_Vid)
{
	int i;
	for( i = 0; i < NUM_DEBUG_PICTURES;i++)
	{
		if(0 != Pictures[i])
		{
			free_mem2Dpel( Pictures[i]);
			Pictures[i] = 0;
		}
	}
	free_mem2Dpel( MskBlock128);
	free_mem2Dpel( MskBlock255);
	free_mem2Dpel( pR);
	free_mem2Dpel( pG);
	free_mem2Dpel( pB);
}

void db_start_frame( VideoParameters* p_Vid)
{
	int co_num, i;
	InputParameters *p_Inp = p_Vid->p_Inp;
  co_num = GetVOIdx(p_Inp,p_Vid->view_id);
  if( (co_num > 0) && (p_Vid->is_depth == 0))
  {
		frame_enable_flag = 1;
		for( i = 0; i < NUM_DEBUG_PICTURES;i++)
			if(0 != Pictures[i])
				memset(&(Pictures[i][0][0]), 0, p_Vid->height*p_Vid->width*sizeof(imgpel));
	}
	else
		frame_enable_flag = 0;
}

void db_set_pred(Macroblock *currMB, short    mode)
{
	Slice *currSlice = currMB->p_Slice;
	VideoParameters *p_Vid = currMB->p_Vid;
	imgpel** pic_pure_pred = 0;
	imgpel** pic_alc_pred = 0;
	imgpel** pic_alcbest_pred = 0;
	int num_pic = mode+NUM_ALC_PRED;

	if(frame_enable_flag == 0)
		return;

	if( mode == PVSP_SKIP)
		num_pic = NUM_MERL_ALC_PRED;
	else if( mode == P8x8)
		num_pic = NUM_8x8_ALC_PRED;

	pic_pure_pred = Pictures[num_pic+QUANTITY_PARTITIONS];
	pic_alc_pred = Pictures[num_pic];

	copy_image_data_16x16(&pic_pure_pred[currMB->pix_y], *currSlice->mb_pred_pure, currMB->pix_x, 0);
	copy_image_data_16x16(&pic_alc_pred[currMB->pix_y], *currSlice->mb_pred, currMB->pix_x, 0);

	if( 0 == mode)
	{
		copy_image_data_16x16(&Pictures[NUM_VSPSKIP_ZONE_MAP][currMB->pix_y], MskBlock255, currMB->pix_x, 0);
	}
}

void db_set_final_mb(Macroblock *currMB)
{
	Slice *currSlice = currMB->p_Slice;
	VideoParameters *p_Vid = currMB->p_Vid;

	if(frame_enable_flag == 0)
		return;

	if(currMB->mb_type == 0)
		copy_image_data_16x16(&Pictures[NUM_SKIP_ZONE_MAP][currMB->pix_y], MskBlock255, currMB->pix_x, 0);
	else if(currMB->mb_type == PVSP_SKIP)
		copy_image_data_16x16(&Pictures[NUM_SKIP_ZONE_MAP][currMB->pix_y], MskBlock128, currMB->pix_x, 0);

	if(currMB->is_alc_mode == 1)
		copy_image_data_16x16(&Pictures[NUM_ALC_ZONE_MAP][currMB->pix_y], MskBlock255, currMB->pix_x, 0);
}

void db_end_frame( VideoParameters* p_Vid)
{
	int i;
  if(frame_enable_flag != 0)
  {
    char fn[256];
    int gag;
    double psnr;
    InputParameters *p_Inp = p_Vid->p_Inp;
    int co_num = GetVOIdx(p_Inp,p_Vid->view_id);

    sprintf( fn, "%s_v%d_%03d.bmp", "bmp/cur_frame", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, p_Vid->imgData.frm_data[0], p_Vid->imgData.frm_data[0], p_Vid->imgData.frm_data[0], p_Vid->height, p_Vid->width);
    sprintf( fn, "%s_v%d_%03d.bmp", "bmp/syn_frame", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, p_Vid->p_vs_pictures[co_num]->imgY,p_Vid->p_vs_pictures[co_num]->imgY, p_Vid->p_vs_pictures[co_num]->imgY, p_Vid->height, p_Vid->width);


		printf("ALC  pred: ");
    i = NUM_SKIP_ALC_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_skip_alc_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], Pictures[NUM_VSPSKIP_ZONE_MAP], NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("Skip: %4.2fdB ", psnr);
    sprintf( fn, "%s_v%d_%03d.bmp", "bmp/map_vspskip", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn,Pictures[NUM_VSPSKIP_ZONE_MAP], Pictures[NUM_VSPSKIP_ZONE_MAP], Pictures[NUM_VSPSKIP_ZONE_MAP], p_Vid->height, p_Vid->width);

    i = NUM_16x16_ALC_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_16x16_alc_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("16x16: %4.2fdB ", psnr);

    i = NUM_16x8_ALC_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_16x8_alc_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("16x8: %4.2fdB ", psnr);

    i = NUM_8x16_ALC_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_8x16_alc_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("8x16: %4.2fdB ", psnr);

    i = NUM_8x8_ALC_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_8x8_alc_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("8x8: %4.2fdB ", psnr);

    i = NUM_MERL_ALC_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_MERL_alc_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("MERL: %4.2fdB", psnr);
    GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], Pictures[NUM_VSPSKIP_ZONE_MAP], NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("(%4.2fdB) ", psnr);

		printf("\n");
		printf("Pure pred: ");

    i = NUM_SKIP_PURE_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_skip_pure_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], Pictures[NUM_VSPSKIP_ZONE_MAP], NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		//GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], Pictures[NUM_VSPSKIP_ZONE_MAP], NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("Skip: %4.2fdB ", psnr);
	
    i = NUM_16x16_PURE_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_16x16_pure_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
    //GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("16x16: %4.2fdB ", psnr);

    i = NUM_16x8_PURE_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_16x8_pure_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
    //GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("16x8: %4.2fdB ", psnr);

    i = NUM_8x16_PURE_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_8x16_pure_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
    //GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("8x16: %4.2fdB ", psnr);

    i = NUM_8x8_PURE_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_8x8_pure_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
    //GetPSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("8x8: %4.2fdB ", psnr);

    i = NUM_MERL_PURE_PRED; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/Mode_MERL_pure_pred", p_Vid->view_id, p_Vid->frame_num);
    save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], NULL, NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("MERL: %4.2fdB", psnr);
    GetMR4x4PSNRwith2mask( p_Vid->imgData.frm_data[0], Pictures[i], Pictures[NUM_VSPSKIP_ZONE_MAP], NULL, 0, 0, p_Vid->height, p_Vid->width, &psnr, &gag);
		printf("(%4.2fdB) ", psnr);

		save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);
		db_set_diff_frame( p_Vid->imgData.frm_data[0], Pictures[NUM_16x16_PURE_PRED] , Pictures[NUM_16x16_ALC_PRED], p_Vid->height, p_Vid->width);
		sprintf( fn, "%s_v%d_%03d.bmp", "bmp/diff_pure_alc_16x16", p_Vid->view_id, p_Vid->frame_num);
		save_pict_imgpel( fn, pR, pG, pB, p_Vid->height, p_Vid->width);

		i = NUM_SKIP_ZONE_MAP; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/map_skip", p_Vid->view_id, p_Vid->frame_num);
		save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);

		i = NUM_ALC_ZONE_MAP; sprintf( fn, "%s_v%d_%03d.bmp", "bmp/map_alc", p_Vid->view_id, p_Vid->frame_num);
		save_pict_imgpel( fn, Pictures[i], Pictures[i], Pictures[i], p_Vid->height, p_Vid->width);

		printf("\n");
  }
	frame_enable_flag = 0;
}

void db_set_diff_frame( imgpel** pFrameSrc, imgpel** pLuma1, imgpel** pLuma2, int h, int w)
{
	int bs = 16;
	int diff,i,j,k,m;
	imgpel R,G,B;
	double uiPSNR_Skip = 0;
	double uiPSNR_Inter = 0;
	double uiPSNR_Intra = 0;
	int iDiff = 0;
	int c1=0;
	int c2=0;
	int c3=0;

#ifdef USEPREDICTIONMODES
        sprintf(tempname,("PredictDecoderMode_v%d_f%05d.frame.yuv" ), viewnum, k1);
        ftemp = fopen(tempname, "rb");
        fread( pModes, 1, h*w, ftemp);
        fclose(ftemp);
#endif
#ifdef USEOCCLLUSIONSMAP
        sprintf(tempname,("Occl_Map_v%d_f%05d.frame.yuv" ), viewnum, k1);
        ftemp = fopen(tempname, "rb");
        fread( pOccls, 1, h*w, ftemp);
        fclose(ftemp);
#endif


        for( i = 0; i < h; i+=bs)
				{
            for( j = 0; j < w; j+=bs)
            {
                double ssd1 = 0;
                double ssd2 = 0;
                int occlcnt = 0;
                R=G=B=0;

#ifdef NOAVER
                int msrc = 0;
                int m1 = 0;
                int m2 = 0;
                for( k = i; k < i+bs; k++)
								{
                    for( m = j; m < j+bs; m++)
                    {
                        m1+=pLuma1[k][m];
                        m2+=pLuma2[k][m];
                        msrc+=pFrameSrc[k][m];
                    }
								}
                m1=(m1+bs*bs/2)/(bs*bs);
                m2=(m2+bs*bs/2)/(bs*bs);
                msrc=(msrc+bs*bs/2)/(bs*bs);

                int delta1 = m1-msrc;
                int delta2 = m2-msrc;
                for( k = i; k < i+bs; k++)
								{
                    for( m = j; m < j+bs; m++)
                    {
                        ssd1+=((int)pFrameSrc[k][m] - (int)pLuma1[k][m]+delta1)*((int)pFrameSrc[k][m] - (int)pLuma1[k][m]+delta1);
                        ssd2+=((int)pFrameSrc[k][m] - (int)pLuma2[k][m]+delta2)*((int)pFrameSrc[k][m] - (int)pLuma2[k][m]+delta2);
                        #ifdef USEOCCLLUSIONSMAP
                            if(pOccls[k*w+m] == 0)// in case of occlusion
                                ++occlcnt;
                        #endif
                    }
								}
#else
                for( k = i; k < i+bs; k++)
								{
                    for( m = j; m < j+bs; m++)
                    {
                        ssd1+=((int)pFrameSrc[k][m] - (int)pLuma1[k][m])*((int)pFrameSrc[k][m] - (int)pLuma1[k][m]);
                        ssd2+=((int)pFrameSrc[k][m] - (int)pLuma2[k][m])*((int)pFrameSrc[k][m] - (int)pLuma2[k][m]);
                        #ifdef USEOCCLLUSIONSMAP
                            if(pOccls[k*w+m] == 0)// in case of occlusion
                                ++occlcnt;
                        #endif
                   }
								}
#endif

                ssd1 /= bs*bs;
                ssd2 /= bs*bs;
                diff = 10*(int)floor(ssd2-ssd1 + 0.5);


                if(abs(diff) <= 10*THRESHOLD)
                {// unimportant area
                    R=G=B=128;
                }
                else if( diff > 0)
                {// block in first frame is better
                    // fill from blue to red
                    G = B = 0; R = diff;
                    if(diff <= 255)
                        B = 255 - diff;
                    else
                        R = 255;
                }
                else if( diff < 0)
                {// block in second frame is better
                    // fill from green to yellow
                    diff = -diff;
                    G = 255; B = 0;
                    if(diff <= 255)
                        R = diff;
                    else
                        R = 255;
                }
                FillRectangle( pR, pG, pB, h, w,  j, i, bs, bs, R, G, B);
                #ifdef USEOCCLLUSIONSMAP
                    if(occlcnt > 2*bs)
                        FillRectangle( pR, pG, pB, h, w,  j+bs/2-1, i+bs/2-1, 2, 2, 0, 0, 0);
                    if(pOccls[k*w+m] == 0)// in case of occlusion
                            continue;
                #endif
#ifdef USEPREDICTIONMODES
                        switch(pEncMode[i*w+j])
#else
               switch(0)
#endif//
               {
               case 0:
                   uiPSNR_Skip   += iDiff * iDiff;
                   c1++;
                   break;
               case 1:
               case 2:
               case 3:
               case 4:
               case 5:
                   uiPSNR_Inter   += iDiff * iDiff;
                   c2++;
                   break;
               default:
                   uiPSNR_Intra   += iDiff * iDiff;
                   c3++;
                   break;
               }
						}
				}
/*
                for( k = i; k < i+bs; k++)
                    for( m = j; m < j+bs; m++)
                        if(pOccls[k][m] == 0)// in case of occlusion
                            pR[k][m] = pG[k][m] = pB[k][m] = 0;
*/
//        TCHAR tempname1[255];
//        _stprintf_s(tempname1, _TEXT("Miracle_%02d_%03d.bmp" ), viewnum, k1);
//        cbRes.SaveWithClip(tempname1);
//        cbRes.Destroy();
}

void GetPSNRwith2mask( imgpel** pPelOrig, imgpel** pPelRec, imgpel** pMaskMap1, imgpel** pMaskMap2, imgpel sig1, imgpel sig2, int h, int w, double* psnr, int* ProcessedNumber)
{
    //===== calculate PSNR =====
    int     iStride   = w;
    int     iWidth    = w;
    int     iHeight   = h;
    int     count	  = 0;
    __int64    uiSSDY    = 0;
    int     x, y;
    int iDiff;
    double fRefValueY;

    int border = 0;
  
    for( y = border; y < iHeight-border; y++ )
    {
        for( x = border; x < iWidth-border; x++ )
        {
            int msk1 = 1;
            int msk2 = 1;
            if(pMaskMap1)
                msk1 = sig1 != pMaskMap1[y][x] ? 1 : 0;
            if(NULL != pMaskMap2)
                msk2 = sig2 != pMaskMap2[y][x] ? 1 : 0;
            if ( (1 == msk1) && (1 == msk2))
            {
              iDiff = (int)pPelOrig[y][x] - (int)pPelRec[y][x];
              uiSSDY   += iDiff * iDiff;
              ++count;
            }
        }
    }

    fRefValueY = 255.0 * 255.0 * count;
    *psnr         = ( uiSSDY ? 10.0 * log10( fRefValueY / (double)uiSSDY ) : 99.99 );
    *ProcessedNumber = count;
    return;
}

void GetMR4x4PSNRwith2mask( imgpel** pPelOrig, imgpel** pPelRec, imgpel** pMaskMap1, imgpel** pMaskMap2, imgpel sig1, imgpel sig2, int h, int w, double* psnr, int* ProcessedNumber)
{
    //===== calculate PSNR =====
    int     iStride   = w;
    int     iWidth    = w;
    int     iHeight   = h;
    int     count	  = 0;
    __int64    uiSSDY    = 0;
    int     x, y,i,j;
    int iDiff;
    double fRefValueY;

    int border = 0;

		for( y = border; y < iHeight-border; y+=4 )
		{
				for( x = border; x < iWidth-border; x+=4 )
				{
						int meanO = 0;
						int meanR = 0;
						int cnt=0;
						for( i = y; i < y+4; i++ )
								for( j = x; j < x+4; j++ )
								{
									int msk1 = 1;
									int msk2 = 1;
									if(pMaskMap1)
											msk1 = sig1 != pMaskMap1[y][x] ? 1 : 0;
									if(NULL != pMaskMap2)
											msk2 = sig2 != pMaskMap2[y][x] ? 1 : 0;
									if ( (1 == msk1) && (1 == msk2))
									{
										meanO+=pPelOrig[i][j];
										meanR+=pPelRec[i][j];
										++cnt;
									}
								}
						if( 0 == cnt)
							continue;

						meanO = (meanO+cnt/2)/cnt;
						meanR = (meanR+cnt/2)/cnt;
						for( i = x; i < x+4; i++ )
						{
								for( j = y; j < y+4; j++ )
								{
									int msk1 = 1;
									int msk2 = 1;
									if(pMaskMap1)
											msk1 = sig1 != pMaskMap1[y][x] ? 1 : 0;
									if(NULL != pMaskMap2)
											msk2 = sig2 != pMaskMap2[y][x] ? 1 : 0;
									if ( (1 == msk1) && (1 == msk2))
									{
										iDiff = ((int)pPelOrig[y][x] - (int)pPelRec[y][x] - meanO + meanR);
										uiSSDY   += iDiff * iDiff;
										++count;
									}
								}
						}
				}
		}

    fRefValueY = 255.0 * 255.0 * count;
    *psnr         = ( uiSSDY ? 10.0 * log10( fRefValueY / (double)uiSSDY ) : 99.99 );
    *ProcessedNumber = count;
    return;
}


int save_pict_imgpel(const char* filename, imgpel **pRImage, imgpel **pGImage, imgpel **pBImage, int height, int width)
{
    int result = 0;
    FILE* pOutputFile = 0;
    BITMAPFILEHEADER BmpFileHead = {0};
    BITMAPINFOHEADER BmpInfoHead = {0};
    int i, j, k;
    unsigned char* pBuffer = 0;
    int BufferAddition = 0;
    int BufferSize     = 0;

#if!(BMP_LOG)
    return 0;
#endif

    do
    {
        if (0 == filename ||
            0 == pRImage  ||
            0 == pGImage  ||
            0 == pBImage  ||
            height <= 0   ||
            width  <= 0)
        {
            _ASSERTE(!"save_pict_imgpel(): wrong input parameter(s).");
            result = (-1);
            break;
        }

        // create internal buffer
        BufferAddition  = ( 4 - ((width*3) & 0x3) ) & 0x3;

        BufferSize  = (width*3 + BufferAddition)*height;

        pBuffer = (unsigned char*)malloc((size_t)BufferSize);

        if (NULL == pBuffer)
        {
            _ASSERTE(!"save_pict_imgpel(): could not allocate enough memory for internal data.");
            result = (-2);
            break;
        }

        memset((void*)pBuffer, 0, BufferSize);

        if (0 == (pOutputFile = fopen(filename, "wb")))
        {
            _ASSERTE(!"save_pict_imgpel(): could not create specified input file.");
            result = (-3);
            break;
        }

        // prepare and write DIB header structures
        BmpFileHead.bfType      = 0x4d42; // BMP
        BmpFileHead.bfSize      = sizeof(BITMAPFILEHEADER) +
                                  sizeof(BITMAPINFOHEADER) +
                                  BufferSize;

        BmpFileHead.bfReserved1 = 0;
        BmpFileHead.bfReserved2 = 0;
        BmpFileHead.bfOffBits   = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

        if (1 != fwrite((void*)&BmpFileHead, sizeof(BITMAPFILEHEADER), 1, pOutputFile))
        {
            _ASSERTE(!"save_pict_imgpel(): could not write DIB file header.");
            result = (-3);
            break;
        }

        BmpInfoHead.biSize          = sizeof(BITMAPINFOHEADER);
        BmpInfoHead.biHeight        = height;
        BmpInfoHead.biWidth         = width;
        BmpInfoHead.biPlanes        = 1;
        BmpInfoHead.biBitCount      = 24;
        BmpInfoHead.biCompression   = BI_RGB;
        BmpInfoHead.biClrUsed       = 0;
        BmpInfoHead.biClrImportant  = 0;
        BmpInfoHead.biXPelsPerMeter = 0;
        BmpInfoHead.biYPelsPerMeter = 0;
        BmpInfoHead.biSizeImage     = 0;

        if (1 != fwrite((void*)&BmpInfoHead, sizeof(BITMAPINFOHEADER), 1, pOutputFile))
        {
            _ASSERTE(!"save_pict_imgpel(): could not write DIB info header.");
            result = (-3);
            break;
        }

        // prepare raw DIB data
        for (i = height -1, k = 0; i >= 0; --i)
        {
            for(j = 0; j < width; ++j)
            {
                pBuffer[k++] = pBImage[i][j];
                pBuffer[k++] = pGImage[i][j];
                pBuffer[k++] = pRImage[i][j];
            }
            k += BufferAddition;
        }

        // write internal data to output file
        if (BufferSize != fwrite((void*)pBuffer, sizeof(unsigned char), BufferSize, pOutputFile))
        {
            _ASSERTE(!"save_pict_imgpel(): could not write DIB info header.");
            result = (-3);
            break;
        }

        fclose(pOutputFile);
        pOutputFile = 0;

        free((void*)pBuffer);
        pBuffer = 0;
        BufferAddition = 0;
        BufferSize     = 0;

        return 0;
    }
    while(0);

    // error handling
    if (0 != pBuffer)
    {
        free((void*)pBuffer);
        pBuffer = 0;
        BufferAddition = 0;
        BufferSize     = 0;
    }

    if (0 != pOutputFile)
    {
        fclose(pOutputFile);
        pOutputFile = 0;
    }
    return result;
}
/*
int save_pict_hf_hifpel(const char* filename, hifpel **pImage, int height, int width)
{
  //_stprintf_s(szFilename, _MAX_PATH, _TEXT("%sPred_PureCFAP_%03d.bmp" ), szPrefix, DispNum);

    int result = 0;

    FILE* pOutputFile = 0;

    BITMAPFILEHEADER BmpFileHead = {0};
    BITMAPINFOHEADER BmpInfoHead = {0};

    int i, j, k;

    unsigned char* pBuffer = 0;
    int BufferAddition = 0;
    int BufferSize     = 0;

#if!(BMP_LOG)
    return 0;
#endif

    do
    {
        if (0 == filename ||
            0 == pImage  ||
            height <= 0   ||
            width  <= 0)
        {
            _ASSERTE(!"save_pict_imgpel(): wrong input parameter(s).");
            result = (-1);
            break;
        }

        // create internal buffer
        BufferAddition  = ( 4 - ((width*3) & 0x3) ) & 0x3;

        BufferSize  = (width*3 + BufferAddition)*height;

        pBuffer = (unsigned char*)malloc((size_t)BufferSize);

        if (NULL == pBuffer)
        {
            _ASSERTE(!"save_pict_imgpel(): could not allocate enough memory for internal data.");
            result = (-2);
            break;
        }

        memset((void*)pBuffer, 0, BufferSize);

        if (0 == (pOutputFile = fopen(filename, "wb")))
        {
            _ASSERTE(!"save_pict_imgpel(): could not create specified input file.");
            result = (-3);
            break;
        }

        // prepare and write DIB header structures
        BmpFileHead.bfType      = 0x4d42; // BMP
        BmpFileHead.bfSize      = sizeof(BITMAPFILEHEADER) +
                                  sizeof(BITMAPINFOHEADER) +
                                  BufferSize;

        BmpFileHead.bfReserved1 = 0;
        BmpFileHead.bfReserved2 = 0;
        BmpFileHead.bfOffBits   = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

        if (1 != fwrite((void*)&BmpFileHead, sizeof(BITMAPFILEHEADER), 1, pOutputFile))
        {
            _ASSERTE(!"save_pict_imgpel(): could not write DIB file header.");
            result = (-3);
            break;
        }

        BmpInfoHead.biSize          = sizeof(BITMAPINFOHEADER);
        BmpInfoHead.biHeight        = height;
        BmpInfoHead.biWidth         = width;
        BmpInfoHead.biPlanes        = 1;
        BmpInfoHead.biBitCount      = 24;
        BmpInfoHead.biCompression   = BI_RGB;
        BmpInfoHead.biClrUsed       = 0;
        BmpInfoHead.biClrImportant  = 0;
        BmpInfoHead.biXPelsPerMeter = 0;
        BmpInfoHead.biYPelsPerMeter = 0;
        BmpInfoHead.biSizeImage     = 0;

        if (1 != fwrite((void*)&BmpInfoHead, sizeof(BITMAPINFOHEADER), 1, pOutputFile))
        {
            _ASSERTE(!"save_pict_imgpel(): could not write DIB info header.");
            result = (-3);
            break;
        }

        // prepare raw DIB data
        for (i = height -1, k = 0; i >= 0; --i)
        {
            for(j = 0; j < width; ++j)
            {
                //pBuffer[k++] = pRImage[i][j];
                //pBuffer[k++] = pGImage[i][j];
                //pBuffer[k++] = pBImage[i][j];
              if(pImage[i][j] == 0)
              {
                pBuffer[k++] = 128;
                pBuffer[k++] = 128;
                pBuffer[k++] = 128;
              }
              else if(pImage[i][j] < 0)
              {
                pBuffer[k++] = 0;
                pBuffer[k++] = iClip3( 0, 255, -2*pImage[i][j]);
                pBuffer[k++] = iClip3( 0, 255, -2*pImage[i][j]);
              }
              else
              {
                pBuffer[k++] = iClip3( 0, 255, 2*pImage[i][j]);
                pBuffer[k++] = iClip3( 0, 255, 2*pImage[i][j]);
                pBuffer[k++] = 0;
              }
            }
            k += BufferAddition;
        }

        // write internal data to output file
        if (BufferSize != fwrite((void*)pBuffer, sizeof(unsigned char), BufferSize, pOutputFile))
        {
            _ASSERTE(!"save_pict_imgpel(): could not write DIB info header.");
            result = (-3);
            break;
        }

        fclose(pOutputFile);
        pOutputFile = 0;

        free((void*)pBuffer);
        pBuffer = 0;
        BufferAddition = 0;
        BufferSize     = 0;

        return 0;
    }
    while(0);

    // error handling
    if (0 != pBuffer)
    {
        free((void*)pBuffer);
        pBuffer = 0;
        BufferAddition = 0;
        BufferSize     = 0;
    }

    if (0 != pOutputFile)
    {
        fclose(pOutputFile);
        pOutputFile = 0;
    }
    return result;
}
*/

// get block form int-pel frame
void GetBlockFromPelArray( imgpel** pFrame, int y_frm, int x_frm, imgpel** pBlock, int blk_h, int blk_w)
{
    int i,j;
    for(i = 0; i < blk_h; i++)
        for(j = 0; j < blk_w; j++)
            pBlock[i][j] = pFrame[(y_frm+i)][x_frm+j];
}

// calculate SAD between templates without DC
int Template1LSADnDC( imgpel** pS1, imgpel** pS2, int subblk_h , int subblk_w, int tmpl_w, int* deltares)
{
    int i,j;
    int sad = 0;
    int m1 = 0;
    int m2 = 0;
    int cnt = 0;
    int wes = 0;
    int delta;
    int allwes = 0;

    for(i = 0; i < subblk_h+tmpl_w; i++)
    {
        for(j = 0; j < subblk_w+tmpl_w; j++)
        {
            if( (i < tmpl_w) || (j < tmpl_w))
            {
                if((i < tmpl_w) && (j < tmpl_w))
                {
                    if(i < j)
                        wes = i+1;
                    else
                        wes = j+1;
                }
                else if(i < tmpl_w)
                    wes = i+1;
                else
                    wes = j+1;

                m1 += wes*pS1[i][j];
                m2 += wes*pS2[i][j];
                cnt += wes;
            }
        }
    }

    delta = (m2-m1 + (cnt>>1))/cnt;

    for(i = 0; i < subblk_h+tmpl_w; i++)
    {
        for(j = 0; j < subblk_w+tmpl_w; j++)
        {
            if( (i < tmpl_w) || (j < tmpl_w))
            {
                if((i < tmpl_w) && (j < tmpl_w))
                {
                    if(i < j)
                        wes = i+1;
                    else
                        wes = j+1;
                }
                else if(i < tmpl_w)
                    wes = i+1;
                else
                    wes = j+1;
                sad += wes*abs((int)pS1[i][j] - (int)pS2[i][j] + delta);
                allwes += wes;
            }
        }
    }

    *deltares = delta;
    return ((sad+(allwes>>1))/allwes)*allwes;
}

// add block with specified value
void AddNum( imgpel** pS1, int chisla, imgpel** pS2, int h , int w)
{
    int i,j;
    for(i = 0; i < h; i++)
    {
        for(j = 0; j < w; j++)
        {
            int res = (int)pS1[i][j]+chisla;
            if(res < 0)
                pS2[i][j] = 0;
            else if(res > 255)
                pS2[i][j] = 255;
            else
              pS2[i][j] = res;
        }
    }
}

void FillRectangle( imgpel** pR, imgpel** pG, imgpel** pB, int height,int width,
                    int s_x, int s_y, int dx, int dy, imgpel R, imgpel G, imgpel B)
{
    int i,j;
    for( i = s_x; i < s_x+dx; i++)
    {
        for( j = s_y; j < s_y+dy; j++)
        {
            pR[ j][i] = R;
            pG[ j][i] = G;
            pB[ j][i] = B;
        }
    }
}

#endif//DEBUG_PROCESSING
