
/*!
*************************************************************************************
* \file me_fullsearch.c
*
* \brief
*    Motion Estimation using Fullsearch
*
* \author
*    Main contributors (see contributors.h for copyright, address and affiliation details)
*      - Alexis Michael Tourapis <alexismt@ieee.org>
*      - Athanasios Leontaris    <aleon@dolby.com>
*
*************************************************************************************
*/

// Includes
#include "contributors.h"
#include <limits.h>

#include "global.h"
#include "image.h"
#include "memalloc.h"
#include "mb_access.h"
#include "refbuf.h"
#include "macroblock.h"
#include "me_distortion.h"
#include "me_fullsearch.h"
#include "mv_search.h"

// Functions
/*!
 ***********************************************************************
 * \brief
 *    Full pixel block motion search
 ***********************************************************************
 */
distblk                                                //  ==> minimum motion cost after search
FullPelBlockMotionSearch (Macroblock   *currMB ,       // <--  current Macroblock
                          MotionVector *pred_mv,       // <--  motion vector predictor in sub-pel units
                          MEBlock      *mv_block,      // <--  motion estimation structure
                          distblk       min_mcost,     // <--  minimum motion cost (cost for center or huge value)
                          int           lambda_factor  // <--  lagrangian parameter for determining motion cost
                          )
{
  VideoParameters *p_Vid = currMB->p_Vid;
  InputParameters *p_Inp = currMB->p_Inp;
  Slice *currSlice = currMB->p_Slice;
  int  search_range = imin(mv_block->searchRange.max_x, mv_block->searchRange.max_y)>> 2;
  distblk mcost;
  int   pos;
  MotionVector cand, center, pred;
  short ref = mv_block->ref_idx;
#if (SMRC_3DV)
  int lamda_mult = LAMBDA_FACTOR(1);
#endif

  StorablePicture *ref_picture = currSlice->listX[mv_block->list+currMB->list_offset][ref];

  int   best_pos      = 0;                                        // position with minimum motion cost
  int   max_pos       = (2*search_range+1)*(2*search_range+1);    // number of search positions

  MotionVector *mv    = &mv_block->mv[(short) mv_block->list];
  int   check_for_00  = (mv_block->blocktype==1 && !p_Inp->rdopt && currSlice->slice_type!=B_SLICE && ref==0);

#if (SMRC_3DV)
	int overlap=0;
	//if( 1 == currMB->is_alc_mode)
	//{
		//if(abs(pred_mv->mv_x) > mv_block->searchRange.max_x || abs(pred_mv->mv_y) > mv_block->searchRange.max_y)
		//{
		//	mv->mv_x = mv->mv_y = 0;
		//	overlap = 1;
		//}
	//	mv->mv_x = mv->mv_y = 0;
	//}
#endif

  center.mv_x      = mv_block->pos_x_padded + mv->mv_x;                        // center position x (in pel units)
  center.mv_y      = mv_block->pos_y_padded + mv->mv_y;                        // center position y (in pel units)
  pred.mv_x        = mv_block->pos_x_padded + pred_mv->mv_x;       // predicted position x (in sub-pel units)
  pred.mv_y        = mv_block->pos_y_padded + pred_mv->mv_y;       // predicted position y (in sub-pel units)
  
#if (SMRC_3DV)
  if( currMB->is_alc_mode)
  {
      search_range = imax(mv_block->searchRange.max_x, mv_block->searchRange.max_y); // in QPEL
      max_pos       = (2*search_range+1)*(2*search_range+1);    // number of search positions

# if (ALC_USE_SSE)
// Lambda recalc should be done for ALC blocks only! May'10'2012
      if (p_Inp->MEErrorMetric[Q_PEL]==ERROR_SAD && currMB->is_alc_mode)
      {
          //convert to the Lambda, that corresponds to SSE metric:
        lambda_factor = lambda_factor * lambda_factor;
          lambda_factor = dist_down(lambda_factor);
      }
#endif
  }
#endif


  //===== loop over all search positions =====
  for (pos=0; pos<max_pos; pos++)
  {
#if (SMRC_3DV)
    if (currMB->is_alc_mode)
    {
        if ( iabs(p_Vid->spiral_qpel_search[pos].mv_x) > mv_block->searchRange.max_x )
          continue;

        if ( iabs(p_Vid->spiral_qpel_search[pos].mv_y) > mv_block->searchRange.max_y )
          continue;
    }
#endif

    //--- set candidate position (absolute position in pel units) ---
    cand.mv_x = center.mv_x + p_Vid->spiral_qpel_search[pos].mv_x;
    cand.mv_y = center.mv_y + p_Vid->spiral_qpel_search[pos].mv_y;

    //--- initialize motion cost (cost for motion vector) and check ---
    mcost = mv_cost (p_Vid, lambda_factor, &cand, &pred);

    if (check_for_00 && cand.mv_x == mv_block->pos_x_padded && cand.mv_y == mv_block->pos_y_padded)
    {
      distblk tmp = weighted_cost (lambda_factor, 16);
      mcost = mcost > tmp? (mcost-tmp): 0;
    }
    if (mcost >= min_mcost)   continue;

    //--- add residual cost to motion cost ---
    mcost += mv_block->computePredFPel(ref_picture, mv_block, min_mcost - mcost, &cand);

    //--- check if motion cost is less than minimum cost ---
    if (mcost < min_mcost)
    {
      best_pos  = pos;
      min_mcost = mcost;
    }
  }


  //===== set best motion vector and return minimum motion cost =====
#if (SMRC_3DV && SMRC_ALC_STFULLSEARCH_AROUND0)
	//if( 1 == currMB->is_alc_mode)
	//{
	//if(1 == overlap)
	//	*mv = p_Vid->spiral_qpel_search[best_pos];
	//else
	//	add_mvs(mv, &p_Vid->spiral_qpel_search[best_pos]);
	//}
	//else
	//	add_mvs(mv, &p_Vid->spiral_qpel_search[best_pos]);
	if( 1 == currMB->is_alc_mode)
		*mv = p_Vid->spiral_qpel_search[best_pos];
	else
	{
		if (best_pos)
		{
			add_mvs(mv, &p_Vid->spiral_qpel_search[best_pos]);
		}
	}
#else
  if (best_pos)
  {
    add_mvs(mv, &p_Vid->spiral_qpel_search[best_pos]);
  }
#endif
  return min_mcost;
}

/*!
 ***********************************************************************
 * \brief
 *    Full pixel block motion search
 ***********************************************************************
 */
distblk                                            //  ==> minimum motion cost after search
FullPelBlockMotionBiPred (Macroblock *currMB,      // <--  current Macroblock
                          int       list,          // <--  reference list
                          MotionVector *pred_mv1,  // <--  motion vector predictor from first list (x|y) in sub-pel units
                          MotionVector *pred_mv2,  // <--  motion vector predictor from second list (x|y) in sub-pel units
                          MotionVector *mv1,       // <--> in: search center (x|y) / out: motion vector (x|y) - in pel units
                          MotionVector *mv2,       // <--> in: search center (x|y) 
                          MEBlock *mv_block,       // <--  motion vector information
                          int       search_range,  // <--  1-d search range in pel units
                          distblk     min_mcost,   // <--  minimum motion cost (cost for center or huge value)
                          int       lambda_factor  // <--  lagrangian parameter for determining motion cost
                          ) 
{
  Slice *currSlice = currMB->p_Slice;
  VideoParameters *p_Vid = currMB->p_Vid;
  short ref = mv_block->ref_idx;
  StorablePicture *ref_picture1 = currSlice->listX[list       + currMB->list_offset][ref];
  StorablePicture *ref_picture2 = currSlice->listX[(list ^ 1) + currMB->list_offset][0];
  distblk mcost;
  int   pos;

  int   best_pos      = 0;                                     // position with minimum motion cost
  int   max_pos       = (2 * (search_range >> 2) +1)*(2 * (search_range >> 2) + 1); // number of search positions

  MotionVector center1, center2, cand, pred1, pred2;
  pred1.mv_x   = mv_block->pos_x_padded + pred_mv1->mv_x; // predicted position x (in sub-pel units)
  pred1.mv_y   = mv_block->pos_y_padded + pred_mv1->mv_y; // predicted position y (in sub-pel units)
  pred2.mv_x   = mv_block->pos_x_padded + pred_mv2->mv_x; // predicted position x (in sub-pel units)
  pred2.mv_y   = mv_block->pos_y_padded + pred_mv2->mv_y; // predicted position y (in sub-pel units)
  center1.mv_x = mv_block->pos_x_padded + mv1->mv_x;              // center position x (in pel units)
  center1.mv_y = mv_block->pos_y_padded + mv1->mv_y;              // center position y (in pel units)
  center2.mv_x = mv_block->pos_x_padded + mv2->mv_x;       // center position x of static mv (in pel units)
  center2.mv_y = mv_block->pos_y_padded + mv2->mv_y;       // center position y of static mv (in pel units)


  //===== loop over all search positions =====
  for (pos=0; pos<max_pos; pos++)
  {
    //--- set candidate position (absolute position in pel units) ---
    cand.mv_x = center1.mv_x + (p_Vid->spiral_search[pos].mv_x << 2);
    cand.mv_y = center1.mv_y + (p_Vid->spiral_search[pos].mv_y << 2);

    //--- initialize motion cost (cost for motion vector) and check ---
    mcost  = mv_cost (p_Vid, lambda_factor, &cand, &pred1);
    mcost += mv_cost (p_Vid, lambda_factor, &center2, &pred2); 

    if (mcost >= min_mcost)   continue;

    //--- add residual cost to motion cost ---
    mcost += mv_block->computeBiPredFPel(ref_picture1, ref_picture2, mv_block, min_mcost - mcost,
      &cand, &center2);

    //--- check if motion cost is less than minimum cost ---
    if (mcost < min_mcost)
    {
      best_pos  = pos;
      min_mcost = mcost;
    }
  }

  //===== set best motion vector and return minimum motion cost =====
  if (best_pos)
  {
    add_mvs(mv1, &p_Vid->spiral_search[best_pos]);
  }
  return min_mcost;
}

/*!
 ***********************************************************************
 * \brief
 *    Sub pixel block motion search
 ***********************************************************************
 */
distblk                                             //  ==> minimum motion cost after search
SubPelBlockMotionSearch (Macroblock   *currMB,      // <--  Current Macroblock
                         MotionVector *pred,        // <--  motion vector predictor in sub-pel units
                         MEBlock      *mv_block,    // <--  motion estimation structure
                         distblk       min_mcost,   // <--  minimum motion cost (cost for center or huge value)
                         int*          lambda       // <--  lagrangian parameter for determining motion cost                         
                         )
{
  Slice *currSlice = currMB->p_Slice;
  VideoParameters *p_Vid = currMB->p_Vid;
  InputParameters *p_Inp = currMB->p_Inp;
  distblk mcost;
  int   pos, best_pos;

  MotionVector cand;
  int list = mv_block->list;
  int cur_list = list + currMB->list_offset;
  short ref = mv_block->ref_idx;
  StorablePicture *ref_picture = currSlice->listX[cur_list][ref];

  MotionVector *mv  = &mv_block->mv[list];

  int   check_position0 = (!p_Inp->rdopt && currSlice->slice_type != B_SLICE && ref==0 && mv_block->blocktype==1 && mv->mv_x == 0 && mv->mv_y ==0);

  int   max_pos2        = ( !p_Vid->start_me_refinement_hp ? imax(1, mv_block->search_pos2) : mv_block->search_pos2);  
  MotionVector cmv;

  int lambda_factor = lambda[H_PEL];

#if (SMRC_3DV && ALC_USE_SSE)
// Lambda recalculation should be done for ALC blocks only! May'10'2012!
  if (p_Inp->MEErrorMetric[Q_PEL]==ERROR_SAD && currMB->is_alc_mode)
       {
      //convert to the Lambda, that corresponds to SSE metric:
            lambda_factor = lambda_factor * lambda_factor;
      lambda_factor = dist_down(lambda_factor);      
        }    
#endif  //(SMRC_3DV && ALC_USE_SSE)

  /*********************************
   *****                       *****
   *****  HALF-PEL REFINEMENT  *****
   *****                       *****
   *********************************/

  //===== loop over search positions =====
  for (best_pos = 0, pos = p_Vid->start_me_refinement_hp; pos < max_pos2; pos++)
  {
    cand.mv_x = mv->mv_x + p_Vid->spiral_hpel_search[pos].mv_x;    // quarter-pel units
    cand.mv_y = mv->mv_y + p_Vid->spiral_hpel_search[pos].mv_y;    // quarter-pel units

    //----- set motion vector cost -----
    mcost = mv_cost (p_Vid, lambda_factor, &cand, pred);


    if (mcost >= min_mcost) continue;

    cmv = pad_MVs(cand, mv_block);

    mcost += mv_block->computePredHPel( ref_picture, mv_block, min_mcost - mcost, &cmv);

    if (pos==0 && check_position0)
    {
      mcost -= weighted_cost (lambda_factor, 16);
    }

    if (mcost < min_mcost)
    {
      min_mcost = mcost;
      best_pos  = pos;
    }
  }
  if (best_pos)
  {
    mv->mv_x = mv->mv_x + p_Vid->spiral_hpel_search[best_pos].mv_x;
    mv->mv_y = mv->mv_y + p_Vid->spiral_hpel_search[best_pos].mv_y;
  }
  if ( !p_Vid->start_me_refinement_qp )
    min_mcost = DISTBLK_MAX;
  /************************************
  *****                          *****
  *****  QUARTER-PEL REFINEMENT  *****
  *****                          *****
  ************************************/

  lambda_factor = lambda[Q_PEL];

#if (SMRC_3DV && ALC_USE_SSE)
 // Lambda recalculation should be done for ALC blocks only! May'10'2012!
  if (p_Inp->MEErrorMetric[Q_PEL]==ERROR_SAD && currMB->is_alc_mode)
        {
      //convert to the Lambda, that corresponds to SSE metric:
            lambda_factor = lambda_factor * lambda_factor;
      lambda_factor = dist_down(lambda_factor);
        }      
#endif //(SMRC_3DV && ALC_USE_SSE)

  //===== loop over search positions =====
  for (best_pos = 0, pos = p_Vid->start_me_refinement_qp; pos < mv_block->search_pos4; pos++)
  {
    cand.mv_x = mv->mv_x + p_Vid->spiral_search[pos].mv_x;    // quarter-pel units
    cand.mv_y = mv->mv_y + p_Vid->spiral_search[pos].mv_y;    // quarter-pel units

    //----- set motion vector cost -----
    mcost = mv_cost (p_Vid, lambda_factor, &cand, pred);

    if (mcost >= min_mcost) continue;

    cmv = pad_MVs(cand, mv_block);

    mcost += mv_block->computePredQPel( ref_picture, mv_block, min_mcost - mcost, &cmv);
    if (mcost < min_mcost)
    {
      min_mcost = mcost;
      best_pos  = pos;
    }
  }
  if (best_pos)
  {
    add_mvs(mv, &p_Vid->spiral_search[best_pos]);
  }

  //===== return minimum motion cost =====
  return min_mcost;
}

/*!
 ***********************************************************************
 * \brief
 *    Sub pixel block motion search
 ***********************************************************************
 */
distblk                                             //  ==> minimum motion cost after search
SubPelBlockMotionSearchHighPrecision (   Macroblock   *currMB,      // <--  Current Macroblock
                                         MotionVector *pred_mv,       // <--  motion vector predictor in sub-pel units
                                         MEBlock      *mv_block,    // <--  motion estimation structure
                                         distblk       min_mcost,   // <--  minimum motion cost (cost for center or huge value)
                                         int*          lambda       // <--  lagrangian parameter for determining motion cost                         
                                         )
{
  VideoParameters *p_Vid = currMB->p_Vid;
  InputParameters *p_Inp = currMB->p_Inp;
  Slice *currSlice = currMB->p_Slice;
  int  search_range = imax(mv_block->searchRange.max_x, mv_block->searchRange.max_y);
  distblk mcost, mvcost, min_mvcost;
  distblk min_error, cur_error;
  int   pos;
  int lambda_factor = lambda[Q_PEL];
  int lambda_factor_alc = lambda_factor;




  MotionVector cand, center, pred;
  short ref = mv_block->ref_idx;

  StorablePicture *ref_picture = currSlice->listX[mv_block->list+currMB->list_offset][ref];
  int   best_pos      = 0;                                        // position with minimum motion cost
  int   max_pos       = (2*search_range+1)*(2*search_range+1);    // number of search positions

  MotionVector *mv    = &mv_block->mv[(short) mv_block->list];
  int   check_for_00  = (mv_block->blocktype==1 && !p_Inp->rdopt && currSlice->slice_type!=B_SLICE && ref==0);
//  center.mv_x      = mv_block->pos_x_padded + mv->mv_x;                        // center position x (in pel units)
//  center.mv_y      = mv_block->pos_y_padded + mv->mv_y;                        // center position y (in pel units)

//#if ((SMRC_3DV) && (ALC_INTERVIEW_ENABLE))
//  int is_virtual_ref_view = currSlice->listX[mv_block->list+currMB->list_offset][ref]->is_virtual_view;
//  int ref_view_id         = currSlice->listX[mv_block->list+currMB->list_offset][ref]->view_id;
//  int is_interview_prediction = (ref_view_id != currSlice->view_id);
//  //coordinates of a block 
//  if (!is_virtual_ref_view && is_interview_prediction)
//  {
//#if (ALC_SET_INTERVIEW_CENTER)
//     center.mv_x = mv_block->pos_x_padded + mv_block->interview_pred_mv.mv_x;        // center position x (in Qpel units)
//     center.mv_y = mv_block->pos_y_padded + mv_block->interview_pred_mv.mv_y;        // center position y (in Qpel units)
//     
//     center.mv_x = (short)iClip3(0, p_Vid->width  << 2, center.mv_x);
//     center.mv_y = (short)iClip3(0, p_Vid->height << 2, center.mv_y);
//#else  // ALC_SET_INTERVIEW_CENTER
//     center.mv_x      = mv_block->pos_x_padded + pred_mv->mv_x;        // center position x (in Qpel units)
//     center.mv_y      = mv_block->pos_y_padded + pred_mv->mv_y;        // center position y (in Qpel units)
//#endif // ALC_SET_INTERVIEW_CENTER
//  }
//  else
//  {
//     center.mv_x      = mv_block->pos_x_padded;                        // center position x (in Qpel units)
//     center.mv_y      = mv_block->pos_y_padded;                        // center position y (in Qpel units)
//  }
//#else  // ((SMRC_3DV) && (ALC_INTERVIEW_ENABLE))
  //coordinates of a block 
  center.mv_x      = mv_block->pos_x_padded;                        // center position x (in Qpel units)
  center.mv_y      = mv_block->pos_y_padded;                        // center position y (in Qpel units)
//#endif // ((SMRC_3DV) && (ALC_INTERVIEW_ENABLE))

  pred.mv_x        = mv_block->pos_x_padded + pred_mv->mv_x;       // predicted position x (in Q-pel units)
  pred.mv_y        = mv_block->pos_y_padded + pred_mv->mv_y;       // predicted position y (in Q-pel units)

  //if (!ref_picture->is_virtual_view)
  //    printf("SubPelBlockMotionSearchHighPrecision: ref view is not virtual!\n");

 # if (ALC_USE_SSE)
  // Lambda recalculation should be done for ALC blocks only! May'10'2012!
  if (p_Inp->MEErrorMetric[Q_PEL]==ERROR_SAD && currMB->is_alc_mode)
  {
      //convert to the Lambda, that corresponds to SSE metric:
      //lambda_factor_alc = lambda_factor * lambda_factor;
      //lambda_factor_alc = dist_down(lambda_factor_alc);
// May'21'2012: More accurate calculation      
      lambda_factor_alc = LAMBDA_FACTOR (p_Vid->lambda_me[currSlice->slice_type][currMB->qp][Q_PEL] * p_Vid->lambda_me[currSlice->slice_type][currMB->qp][Q_PEL]);     


      //LAMBDA_FACTOR(
  }    
#endif
  
  min_mcost = DISTBLK_MAX;

  //===== loop over all search positions =====
  for (pos=0; pos<max_pos; pos++)
  {
    //--- set candidate position (absolute position in pel units) ---
//    cand.mv_x = center.mv_x + p_Vid->spiral_qpel_search[pos].mv_x;
//    cand.mv_y = center.mv_y + p_Vid->spiral_qpel_search[pos].mv_y;


      if ( abs(p_Vid->spiral_search[pos].mv_x) > mv_block->searchRange.max_x )
          continue;

      if ( abs(p_Vid->spiral_search[pos].mv_y) > mv_block->searchRange.max_y )
          continue;

    cand.mv_x = center.mv_x + p_Vid->spiral_search[pos].mv_x;
    cand.mv_y = center.mv_y + p_Vid->spiral_search[pos].mv_y;

    //--- initialize motion cost (cost for motion vector) and check ---
    mvcost = mcost = mv_cost (p_Vid, lambda_factor_alc, &cand, &pred);

    //--- add residual cost to motion cost ---
    cur_error = mv_block->computePredQPel(ref_picture, mv_block, min_mcost - mcost, &cand);
    mcost += cur_error;

    //--- check if motion cost is less than minimum cost ---
    if (mcost < min_mcost)
    {
      best_pos  = pos;
      min_mcost = mcost;
      min_mvcost = mvcost;
      min_error = cur_error;
    }
/*
    else
    if (mcost == min_mcost) // we can do it as we DO NOT USE earlier termination in computePredQPel ( so, if result of computePredQPel is equal to min_mcost - it's true!)
        if (cur_error < min_error)
        {
            best_pos  = pos;            
            min_mvcost = mvcost;
            min_error = cur_error;
        }
*/
  }


# if (ALC_USE_SSE)
  if (p_Inp->MEErrorMetric[Q_PEL]==ERROR_SAD)
  {
      //output cost has to be recalculated due-to here we used SSE metric instead of SAD and Lambda is differed than that of used for SAD
      cand.mv_x = center.mv_x + p_Vid->spiral_search[best_pos].mv_x;
      cand.mv_y = center.mv_y + p_Vid->spiral_search[best_pos].mv_y;

      // cost of MV
      min_mcost = mv_cost (p_Vid, lambda_factor, &cand, &pred);

#if (USE_MR_METRIC_FOR_ALC)
    #if (MRSAD_4x4)
      // MR_SAD by MR_SAD4x4
      min_error = computeMRSAD4x4(ref_picture, mv_block, INT_MAX, &cand);
    #else
      // MR_SAD by MR_SAD4x4
      min_error = computeMRSAD(ref_picture, mv_block, INT_MAX, &cand);
    #endif //MRSAD_4x4
#else
      printf("SubPelBlockMotionSearchHighPrecision: Unsupported!");
#endif //USE_MR_METRIC_FOR_ALC


      min_mcost = min_mcost + min_error;     
  }
#endif //ALC_USE_SSE



  //===== set best motion vector and return minimum motion cost =====
//  if (best_pos)
//  {
    //add_mvs(mv, &p_Vid->spiral_search[best_pos]);
      mv->mv_x = p_Vid->spiral_search[best_pos].mv_x;
      mv->mv_y = p_Vid->spiral_search[best_pos].mv_y;
//  }
  return min_mcost;
}

/*!
***********************************************************************
* \brief
*    Bipred Sub pixel block motion search
***********************************************************************
*/
distblk                                           //  ==> minimum motion cost after search
SubPelBlockSearchBiPred (Macroblock   *currMB,    // <--  current Macroblock
                         MEBlock      *mv_block,  // <--  motion vector information
                         int           list,      // <--  reference picture list
                         MotionVector *pred_mv1,  // <--  motion vector predictor (x) in sub-pel units
                         MotionVector *pred_mv2,  // <--  motion vector predictor (x) in sub-pel units
                         MotionVector *mv1,       // <--> in: search center (x) / out: motion vector (x) - in pel units
                         MotionVector *mv2,       // <--> in: search center (x) / out: motion vector (x) - in pel units
                         distblk       min_mcost, // <--  minimum motion cost (cost for center or huge value)
                         int*          lambda     // <--  lagrangian parameter for determining motion cost
                         )
{
  VideoParameters *p_Vid = currMB->p_Vid;
  Slice *currSlice = currMB->p_Slice; 
  int   list_offset   = p_Vid->mb_data[currMB->mbAddrX].list_offset;
  distblk mcost;
  int   pos, best_pos;
  MotionVector cand;
  int   start_hp    = (min_mcost == DISTBLK_MAX) ? 0 : p_Vid->start_me_refinement_hp;
  int   max_pos2        = ( !p_Vid->start_me_refinement_hp ? imax(1, mv_block->search_pos2) : mv_block->search_pos2);

  MotionVector cmv, smv = pad_MVs(*mv2, mv_block);
  short ref = mv_block->ref_idx;

  StorablePicture *ref_picture1 = currSlice->listX[list       + list_offset][ref];
  StorablePicture *ref_picture2 = currSlice->listX[(list ^ 1) + list_offset][0];

  int lambda_factor = lambda[H_PEL];

  /*********************************
   *****                       *****
   *****  HALF-PEL REFINEMENT  *****
   *****                       *****
   *********************************/

  //===== loop over search positions =====
  for (best_pos = 0, pos = start_hp; pos < max_pos2; pos++)
  {
    cand.mv_x = mv1->mv_x + p_Vid->spiral_hpel_search[pos].mv_x;    // quarter-pel units
    cand.mv_y = mv1->mv_y + p_Vid->spiral_hpel_search[pos].mv_y;    // quarter-pel units

    //----- set motion vector cost -----
    mcost  = mv_cost (p_Vid, lambda_factor, &cand, pred_mv1);
    mcost += mv_cost (p_Vid, lambda_factor,   mv2, pred_mv2);

    if (mcost >= min_mcost) continue;

    cmv = pad_MVs(cand, mv_block);
    mcost += mv_block->computeBiPredHPel(ref_picture1, ref_picture2, mv_block, min_mcost - mcost, &cmv, &smv);

    if (mcost < min_mcost)
    {
      min_mcost = mcost;
      best_pos  = pos;
    }
  }

  if (best_pos)
  {
    add_mvs(mv1, &p_Vid->spiral_hpel_search[best_pos]);
  }

  /************************************
  *****                          *****
  *****  QUARTER-PEL REFINEMENT  *****
  *****                          *****
  ************************************/
  if ( !p_Vid->start_me_refinement_qp )
    min_mcost = DISTBLK_MAX;

  lambda_factor = lambda[Q_PEL];

  //===== loop over search positions =====
  for (best_pos = 0, pos = p_Vid->start_me_refinement_qp; pos < mv_block->search_pos4; pos++)
  {
    cand.mv_x = mv1->mv_x + p_Vid->spiral_search[pos].mv_x;    // quarter-pel units
    cand.mv_y = mv1->mv_y + p_Vid->spiral_search[pos].mv_y;    // quarter-pel units

    //----- set motion vector cost -----
    mcost  = mv_cost (p_Vid, lambda_factor, &cand, pred_mv1);
    mcost += mv_cost (p_Vid, lambda_factor,   mv2, pred_mv2);

    if (mcost >= min_mcost) continue;
    
    cmv = pad_MVs(cand, mv_block);
    mcost += mv_block->computeBiPredQPel(ref_picture1, ref_picture2, mv_block, min_mcost - mcost, &cmv, &smv);

    if (mcost < min_mcost)
    {
      min_mcost = mcost;
      best_pos  = pos;
    }

  }

  if (best_pos)
  {
    add_mvs(mv1, &p_Vid->spiral_search[best_pos]);
  }

  //===== return minimum motion cost =====
  return min_mcost;
}

